"""
Factories to configure Flask.

"""
from flask import Flask

from microcosm.api import defaults
import microcosm.opaque  # noqa


@defaults(
    port=5000,
)
def configure_flask(graph):
    """
    Create the Flask instance (only), bound to the "flask" key.

    Conventions should refer to `graph.flask` to avoid circular dependencies.

    """
    app = Flask(graph.metadata.import_name)
    app.debug = graph.metadata.debug
    app.testing = graph.metadata.testing

    # copy in the graph's configuration for non-nested keys
    app.config.update({
        key: value
        for key, value in graph.config.items()
        if not isinstance(value, dict)
    })

    return app


def configure_flask_app(graph):
    """
    Configure a Flask application with common conventions, bound to the "app" key.

    """
    graph.use(
        "audit",
        "request_context",
        "basic_auth",
        "error_handlers",
        "logger",
        "opaque",
    )
    return graph.flask



"""
Pagination support.

Most applications start with offset/limit pagination because it's simplest to implement,
but other pagination schemes are possible and can be more performant especially in infinite
scroll settings. This module encapsulates paging into a set of extensible, inter-related objects.

 -  A `Page` represents information about a specific page (for example, the currently requested one)
 -  A `PageSchema` defines a (marshmallow) schema for decoding a page from some data (e.g. the query string)
 -  A `PaginatedList` defines a list of items that knows about its current page and *may* define HAL-style
    links to other pages.
 -  A `PaginatedListSchema` defines a (marshmallow) schema for encoding a paginated list (e.g. in a response)


Typical Usage:

    # pre-conditions: we are processes a search operation for `foo`
    ns, operation, foo_schema = Namespace("foo"), Operation.Search, Schema()

    # choose a page implementation
    page_cls = OffsetLimitPage

    # choose a page schema (we may wish to extend the base class for more search arguments)
    page_schema = OffsetLimitPageSchema()

    # construct the current page from the query string
    page = page_cls.from_query_string(page_schema)

    # perform the search; result is typically a tuple including items and a total count
    result = some_search_func(page)

    # transform the result into a paginated list and possibly response headers
    paginated_list, headers = page.to_paginated_list(result, ns, operation)

    # encode the result
    paginated_list_schema = page_cls.make_paginated_list_schema_class(ns, foo_schema)()
    return dump_response_Data(paginated_list_schema, paginated_list, headers=headers)

"""
from marshmallow import fields, Schema
from flask import request

from microcosm_flask.conventions.encoding import encode_count_header, load_query_string_data
from microcosm_flask.linking import Link, Links


def identity(x):
    """
    Identity function.

    """
    return x


# NB: lots of code currently uses `PageSchema` to refer to `OffsetLimitPageSchema`
# keeping this (mis)naming for backwards compatibilty
class PageSchema(Schema):
    offset = fields.Integer(missing=None)
    limit = fields.Integer(missing=None)


class OffsetLimitPageSchema(PageSchema):
    pass


class PaginatedList(object):
    """
    A list of items with knowledge of a page.

    Includes HAL-style links (e.g for the current or next page)

    """
    def __init__(self, items, _page, _ns, _operation, _context):
        self.items = items
        self._page = _page
        self._ns = _ns
        self._operation = _operation
        self._context = _context

    @property
    def _links(self):
        return self.links.to_dict()

    @property
    def links(self):
        """
        Include a self link.

        """
        links = Links()
        links["self"] = Link.for_(
            self._operation,
            self._ns,
            qs=self._page.to_items(),
            **self._context
        )
        return links


class OffsetLimitPaginatedList(PaginatedList):
    """
    A paginated list using offset/limit style paging.

    """
    def __init__(self, items, count, _page, _ns, _operation, _context):
        super(OffsetLimitPaginatedList, self).__init__(
            items=items,
            _page=_page,
            _ns=_ns,
            _operation=_operation,
            _context=_context,
        )
        self.count = count

    @property
    def offset(self):
        return self._page.offset

    @property
    def limit(self):
        return self._page.limit

    @property
    def links(self):
        """
        Include previous and next links.

        """
        links = super(OffsetLimitPaginatedList, self).links
        if self._page.offset + self._page.limit < self.count:
            links["next"] = Link.for_(
                self._operation,
                self._ns,
                qs=self._page.next_page.to_items(),
                **self._context
            )
        if self.offset > 0:
            links["prev"] = Link.for_(
                self._operation,
                self._ns,
                qs=self._page.prev_page.to_items(),
                **self._context
            )
        return links


class Page(object):
    """
    Encapsulates pagination information.

    """
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def to_items(self, func=str):
        """
        Contruct a list of dictionary items.

        The items are normalized using:
          -  A sort function by key (for consistent results)
          -  A transformation function for values

        The transformation function will default to `str`, which is a good choice when encoding values
        as part of a response; this requires that complex types (UUID, Enum, etc.) have a valid string
        encoding.

        The transformation function should be set to `identity` in cases where raw values are desired;
        this is normally necessary when passing page data to controller functions as kwargs.

        """
        return [
            (key, func(self.kwargs[key]))
            for key in sorted(self.kwargs.keys())
        ]

    def to_dict(self, func=str):
        return dict(self.to_items(func=func))

    def to_paginated_list(self, result, _ns, _operation, **kwargs):
        """
        Convert a controller result to a paginated list.

        The result format is assumed to meet the contract of this page class's `parse_result` function.

        """
        items, context = self.parse_result(result)
        headers = dict()
        paginated_list = PaginatedList(
            items=items,
            _page=self,
            _ns=_ns,
            _operation=_operation,
            _context=context,
        )
        return paginated_list, headers

    @classmethod
    def parse_result(cls, result):
        """
        Parse a simple items result.

        May either be two item tuple containing items and a context dictionary (see: relation convention)
        or a list of items.

        """
        if isinstance(result, tuple) == 2:
            items, context = result
        else:
            context = {}
            items = result
        return items, context

    @classmethod
    def from_query_string(cls, schema, qs=None):
        """
        Extract a page from the current query string.

        :param qs: a query string dictionary (`request.args` will be used if omitted)

        """
        dct = load_query_string_data(schema, qs)
        return cls.from_dict(dct)

    @classmethod
    def from_dict(cls, dct):
        return cls(**dct)

    @classmethod
    def make_paginated_list_schema_class(cls, ns, item_schema):
        """
        Generate a schema class that represents a paginted list of items.

        """
        class PaginatedListSchema(Schema):
            __alias__ = "{}_list".format(ns.subject_name)
            items = fields.List(fields.Nested(item_schema), required=True)
            _links = fields.Raw()

        return PaginatedListSchema


class OffsetLimitPage(Page):
    """
    Offset/limit based paging.

    """
    def __init__(self, offset=None, limit=None, **kwargs):
        super(OffsetLimitPage, self).__init__(**kwargs)
        self.offset = self.default_offset if offset is None else offset
        self.limit = self.default_limit if limit is None else limit

    @property
    def next_page(self):
        return OffsetLimitPage(
            offset=self.offset + self.limit,
            limit=self.limit,
            **self.kwargs
        )

    @property
    def prev_page(self):
        return OffsetLimitPage(
            offset=self.offset - self.limit,
            limit=self.limit,
            **self.kwargs
        )

    @property
    def default_offset(self):
        return 0

    @property
    def default_limit(self):
        try:
            return int(request.headers["X-Request-Limit"])
        except:
            return 20

    def to_items(self, func=str):
        return [
            ("offset", self.offset),
            ("limit", self.limit),
        ] + super(OffsetLimitPage, self).to_items(func=func)

    def to_paginated_list(self, result, _ns, _operation, **kwargs):
        items, count, context = self.parse_result(result)
        headers = encode_count_header(count)
        paginated_list = OffsetLimitPaginatedList(
            items=items,
            count=count,
            _page=self,
            _ns=_ns,
            _operation=_operation,
            _context=context,
        )
        return paginated_list, headers

    @classmethod
    def parse_result(cls, result):
        """
        Parse an items + count tuple result.

        May either be three item tuple containing items, count, and a context dictionary (see: relation convention)
        or a two item tuple containing only items and count.

        """
        if len(result) == 3:
            items, count, context = result
        else:
            context = {}
            items, count = result
        return items, count, context

    @classmethod
    def make_paginated_list_schema_class(cls, ns, item_schema):
        class PaginatedListSchema(Schema):
            __alias__ = "{}_list".format(ns.subject_name)

            offset = fields.Integer(required=True)
            limit = fields.Integer(required=True)
            count = fields.Integer(required=True)
            items = fields.List(fields.Nested(item_schema), required=True)
            _links = fields.Raw()

        return PaginatedListSchema



"""
Pull resource definitions from an input source.

"""
from logging import getLogger
from sys import stdin

from requests import get
from yaml import load_all


logger = getLogger("sync.pull")


def iter_links(resource):
    """
    Iterate over links in a resource.

    """
    for relation, value in resource.get("_links", {}).items():
        if isinstance(value, list):
            for link in value:
                yield relation, link
        else:
            yield relation, value


def iter_resources(resource):
    """
    Iterate over resources in a resource.

    """
    if "items" in resource:
        # collection: embeds resources as items
        for embedded in resource.get("items"):
            href = embedded["_links"]["self"]["href"]
            logger.debug("Found embedded resource: {}".format(href))
            yield href, embedded
    else:
        # instance
        href = resource["_links"]["self"]["href"]
        logger.debug("Found resource: {}".format(href))
        yield href, resource


def sort_links(resource):
    """
    Sort resources links for easier diffs.

    """
    for relation, value in resource["_links"].items():
        if isinstance(value, list):
            resource["_links"][relation] = sorted(value, key=lambda value: value["href"])
    return resource


def pull_json(args, base_url):
    """
    Pull JSON resources by spidering a base url.

    """
    exclude_first = args.exclude_first
    stack = [base_url]

    seen = set()
    while stack:
        uri = stack.pop()

        logger.info("Fetching resource URI: {}".format(uri))
        response = get(uri)
        response.raise_for_status()
        data = response.json()

        for href, resource in iter_resources(data):
            if exclude_first:
                # skipping the first resource - if it's a discovery resource - avoids
                # pushing back state that cannot be persisted
                exclude_first = False
                continue
            sort_links(resource)
            for relation, links in iter_links(resource):
                if href not in seen and any(pattern.match(relation) for pattern in args.relation_patterns):
                    seen.add(href)
                    stack.append(href)
            yield href, resource

        for relation, link in iter_links(data):
            href = link["href"]
            # follow top-level search links and pagination next links
            if href not in seen and any(pattern.match(relation) for pattern in args.relation_patterns):
                seen.add(href)
                stack.append(href)


def pull_yaml(args, source):
    """
    Pull YAML resources from a file-like object.

    """
    for dct in load_all(source):
        for item in dct.items():
            yield item


def pull(args):
    logger.info("Pulling resources from: {}".format(args.input))

    if args.input == "-":
        for href, resource in pull_yaml(args, stdin):
            yield href, resource
    elif args.input.startswith("http"):
        for href, resource in pull_json(args, args.input):
            yield href, resource
    else:
        with open(args.input) as file_:
            for href, resource in pull_yaml(args, file_):
                yield href, resource



"""
Generate JSON Schema for Marshmallow schemas.

"""
from logging import getLogger
from six import string_types

from marshmallow import fields

from microcosm_flask.fields import (
    EnumField,
    LanguageField,
    QueryStringList,
    URIField,
)
from microcosm_flask.naming import name_for
from microcosm_flask.swagger.naming import type_name


logger = getLogger("microcosm_flask.swagger")


# see: https://github.com/marshmallow-code/apispec/blob/dev/apispec/ext/marshmallow/swagger.py
FIELD_MAPPINGS = {
    EnumField: (None, None),
    LanguageField: ("string", "language"),
    QueryStringList: ("array", None),
    URIField: ("string", "uri"),
    fields.Boolean: ("boolean", None),
    fields.Date: ("string", "date"),
    fields.DateTime: ("string", "date-time"),
    fields.Decimal: ("number", None),
    fields.Dict: ("object", None),
    fields.Email: ("string", "email"),
    fields.Field: ("object", None),
    fields.Float: ("number", "float"),
    fields.Integer: ("integer", "int32"),
    fields.List: ("array", None),
    fields.Method: ("object", None),
    fields.Nested: (None, None),
    fields.Number: ("number", None),
    fields.Raw: ("object", None),
    fields.String: ("string", None),
    fields.Time: ("string", None),
    fields.URL: ("string", "url"),
    fields.UUID: ("string", "uuid"),
}


SWAGGER_TYPE = "__swagger_type__"
SWAGGER_FORMAT = "__swagger_format__"


def is_int(value):
    try:
        int(value)
    except:
        return False
    else:
        return True


def swagger_field(field, swagger_type="string", swagger_format=None):
    setattr(field, SWAGGER_TYPE, swagger_type)
    setattr(field, SWAGGER_FORMAT, swagger_format)
    return field


def build_parameter(field):
    """
    Build a parameter from a marshmallow field.

    See: https://github.com/marshmallow-code/apispec/blob/dev/apispec/ext/marshmallow/swagger.py#L81

    """
    try:
        field_type, field_format = FIELD_MAPPINGS[type(field)]
    except KeyError:
        if hasattr(field, SWAGGER_TYPE):
            field_type = getattr(field, SWAGGER_TYPE)
            field_format = getattr(field, SWAGGER_FORMAT, None)
        else:
            logger.exception("No mapped swagger type for marshmallow field: {}".format(
                field,
            ))
            raise

    parameter = {}
    if field_type:
        parameter["type"] = field_type
    if field.metadata.get("description"):
        parameter["description"] = field.metadata["description"]
    if field_format:
        parameter["format"] = field_format
    if field.default:
        parameter["default"] = field.default
    # NB: all marshallow Number fields support as_string
    if getattr(field, 'as_string', None):
        parameter["type"] = "string"
        if isinstance(field, fields.Decimal):
            parameter["format"] = "decimal"

    # enums
    enum = getattr(field, "enum", None)
    if enum:
        enum_values = [
            choice.value if field.by_value else choice.name
            for choice in enum
        ]
        if all((isinstance(enum_value, string_types) for enum_value in enum_values)):
            enum_type = "string"
        elif all((is_int(enum_value) for enum_value in enum_values)):
            enum_type = "integer"
        else:
            raise Exception("Cannot infer enum type for field: {}".format(field.name))

        parameter["type"] = enum_type
        parameter["enum"] = enum_values

    # nested
    if isinstance(field, fields.Nested):
        parameter["$ref"] = "#/definitions/{}".format(type_name(name_for(field.schema)))

    # arrays
    if isinstance(field, fields.List):
        parameter["items"] = build_parameter(field.container)

    return parameter


def build_schema(marshmallow_schema):
    """
    Build JSON schema from a marshmallow schema.

    """
    fields = list(iter_fields(marshmallow_schema))
    required_fields = [
        field.dump_to or name
        for name, field in fields
        if field.required and not field.allow_none
    ]
    schema = {
        "type": "object",
        "properties": {
            field.dump_to or name: build_parameter(field)
            for name, field in fields
        }
    }
    if required_fields:
        schema["required"] = required_fields
    return schema


def iter_fields(marshmallow_schema):
    """
    Iterate through marshmallow schema fields.

    Generates: name, field pairs

    """
    for name in sorted(marshmallow_schema.fields.keys()):
        yield name, marshmallow_schema.fields[name]


def iter_schemas(marshmallow_schema):
    """
    Build zero or more JSON schemas for a marshmallow schema.

    Generates: name, schema pairs.

    """
    if not marshmallow_schema:
        return

    base_schema = build_schema(marshmallow_schema)
    base_schema_name = type_name(name_for(marshmallow_schema))
    yield base_schema_name, base_schema

    for name, field in iter_fields(marshmallow_schema):
        if isinstance(field, fields.Nested):
            nested_schema = build_schema(field.schema)
            nested_schema_name = type_name(name_for(field.schema))
            yield nested_schema_name, nested_schema
            for subname, subfield in iter_schemas(field.schema):
                yield subname, subfield
        if isinstance(field, fields.List) and isinstance(field.container, fields.Nested):
            nested_schema = build_schema(field.container.schema)
            nested_schema_name = type_name(name_for(field.container.schema))
            yield nested_schema_name, nested_schema
            for subname, subfield in iter_schemas(field.container.schema):
                yield subname, subfield



"""
Conventions for canonical relation endpoints (mapping one resource to another).

For relations, endpoint definitions require that the `Namespace` contain *both*
a subject and an object.

"""
from inflection import pluralize
from marshmallow import Schema

from microcosm_flask.conventions.base import Convention
from microcosm_flask.conventions.encoding import (
    dump_response_data,
    load_query_string_data,
    load_request_data,
    merge_data,
    require_response_data,
)
from microcosm_flask.conventions.registry import qs, request, response
from microcosm_flask.namespaces import Namespace
from microcosm_flask.operations import Operation
from microcosm_flask.paging import OffsetLimitPage


class RelationConvention(Convention):

    @property
    def page_cls(self):
        return OffsetLimitPage

    def __init__(self, graph):
        super(RelationConvention, self).__init__(graph)

    def configure_createfor(self, ns, definition):
        """
        Register a create-for relation endpoint.

        The definition's func should be a create function, which must:
        - accept kwargs for the new instance creation parameters
        - return the created instance

        :param ns: the namespace
        :param definition: the endpoint definition

        """
        @self.add_route(ns.relation_path, Operation.CreateFor, ns)
        @request(definition.request_schema)
        @response(definition.response_schema)
        def create(**path_data):
            request_data = load_request_data(definition.request_schema)
            response_data = require_response_data(definition.func(**merge_data(path_data, request_data)))
            return dump_response_data(definition.response_schema, response_data, Operation.CreateFor.value.default_code)

        create.__doc__ = "Create a new {} relative to a {}".format(pluralize(ns.object_name), ns.subject_name)

    def configure_deletefor(self, ns, definition):
        """
        Register a delete-for relation endpoint.

        The definition's func should be a delete function, which must:
        - accept kwargs for path data
        - return truthy/falsey

        :param ns: the namespace
        :param definition: the endpoint definition

        """
        @self.add_route(ns.relation_path, Operation.DeleteFor, ns)
        def delete(**path_data):
            require_response_data(definition.func(**path_data))
            return "", Operation.DeleteFor.value.default_code

        delete.__doc__ = "Delete a {} relative to a {}".format(pluralize(ns.object_name), ns.subject_name)

    def configure_replacefor(self, ns, definition):
        """
        Register a replace-for relation endpoint.

        For typical usage, this relation is not strictly required; once an object exists and has its own ID,
        it is better to operate on it directly via dedicated CRUD routes.
        However, in some cases, the composite key of (subject_id, object_id) is required to look up the object.
        This happens, for example, when using DynamoDB where an object which maintains both a hash key and a range key
        requires specifying them both for access.

        The definition's func should be a replace function, which must:

        - accept kwargs for the new instance replacement parameters
        - return the instance

        :param ns: the namespace
        :param definition: the endpoint definition

        """
        @self.add_route(ns.relation_path, Operation.ReplaceFor, ns)
        @request(definition.request_schema)
        @response(definition.response_schema)
        def replace(**path_data):
            request_data = load_request_data(definition.request_schema)
            response_data = require_response_data(definition.func(**merge_data(path_data, request_data)))
            return dump_response_data(
                definition.response_schema,
                response_data,
                Operation.ReplaceFor.value.default_code,
            )

        replace.__doc__ = "Replace a {} relative to a {}".format(pluralize(ns.object_name), ns.subject_name)

    def configure_updatefor(self, ns, definition):
        """
        Register an update-for relation endpoint.

        For typical usage, this relation is not strictly required; once an object exists and has its own ID,
        it is better to operate on it directly via dedicated CRUD routes.
        However, in some cases, the composite key of (subject_id, object_id) is required to look up the object.
        This happens, for example, when using DynamoDB where an object which maintains both a hash key and a range key
        requires specifying them both for access.

        The definition's func should be an update function, which must:

        - accept kwargs for the new instance replacement parameters
        - return the instance

        :param ns: the namespace
        :param definition: the endpoint definition

        """
        @self.add_route(ns.relation_path, Operation.UpdateFor, ns)
        @request(definition.request_schema)
        @response(definition.response_schema)
        def replace(**path_data):
            request_data = load_request_data(definition.request_schema)
            response_data = require_response_data(definition.func(**merge_data(path_data, request_data)))
            return dump_response_data(
                definition.response_schema,
                response_data,
                Operation.UpdateFor.value.default_code,
            )

        replace.__doc__ = "Replace a {} relative to a {}".format(pluralize(ns.object_name), ns.subject_name)

    def configure_retrievefor(self, ns, definition):
        """
        Register a relation endpoint.

        The definition's func should be a retrieve function, which must:
        - accept kwargs for path data and optional request data
        - return an item

        The definition's request_schema will be used to process query string arguments, if any.

        :param ns: the namespace
        :param definition: the endpoint definition

        """
        request_schema = definition.request_schema or Schema()

        @self.add_route(ns.relation_path, Operation.RetrieveFor, ns)
        @qs(request_schema)
        @response(definition.response_schema)
        def retrieve(**path_data):
            request_data = load_query_string_data(request_schema)
            response_data = require_response_data(definition.func(**merge_data(path_data, request_data)))
            return dump_response_data(definition.response_schema, response_data)

        retrieve.__doc__ = "Retrieve {} relative to a {}".format(pluralize(ns.object_name), ns.subject_name)

    def configure_searchfor(self, ns, definition):
        """
        Register a relation endpoint.

        The definition's func should be a search function, which must:
        - accept kwargs for the query string (minimally for pagination)
        - return a tuple of (items, count, context) where count is the total number of items
          available (in the case of pagination) and context is a dictionary providing any
          needed context variables for constructing pagination links

        The definition's request_schema will be used to process query string arguments.

        :param ns: the namespace
        :param definition: the endpoint definition

        """
        paginated_list_schema = self.page_cls.make_paginated_list_schema_class(
            ns.object_ns,
            definition.response_schema,
        )()

        @self.add_route(ns.relation_path, Operation.SearchFor, ns)
        @qs(definition.request_schema)
        @response(paginated_list_schema)
        def search(**path_data):
            request_data = load_query_string_data(definition.request_schema)
            page = self.page_cls.from_query_string(definition.request_schema)
            result = definition.func(**merge_data(path_data, request_data))
            response_data, headers = page.to_paginated_list(result, ns, Operation.SearchFor)
            return dump_response_data(paginated_list_schema, response_data, headers=headers)

        search.__doc__ = "Search for {} relative to a {}".format(pluralize(ns.object_name), ns.subject_name)


def configure_relation(graph, ns, mappings, path_prefix=""):
    """
    Register relation endpoint(s) between two resources.

    """
    ns = Namespace.make(ns, path=path_prefix)
    convention = RelationConvention(graph)
    convention.configure(ns, mappings)



"""
Conventions for aliasing endpoints.

"""
from flask import redirect
from microcosm_flask.conventions.base import Convention
from microcosm_flask.naming import name_for
from microcosm_flask.operations import Operation


class AliasConvention(Convention):

    def configure_alias(self, ns, definition):
        """
        Register an alias endpoint which will redirect to a resource's retrieve endpoint.

        Note that the retrieve endpoint MUST be registered prior to the alias endpoint.

        The definition's func should be a retrieve function, which must:
        - accept kwargs for path data
        - return a resource

        :param ns: the namespace
        :param definition: the endpoint definition

        """
        @self.add_route(ns.alias_path, Operation.Alias, ns)
        def retrieve(**path_data):
            resource = definition.func(**path_data)

            kwargs = dict()
            identifier = "{}_id".format(name_for(ns.subject))
            kwargs[identifier] = resource.id
            url = ns.url_for(Operation.Retrieve, **kwargs)

            return redirect(url)

        retrieve.__doc__ = "Alias a {} by name".format(ns.subject_name)


def configure_alias(graph, ns, mappings):
    """
    Register Alias endpoints for a resource object.

    """
    convention = AliasConvention(graph)
    convention.configure(ns, mappings)



"""
Audit log support for Flask routes.

"""
from collections import namedtuple
from contextlib import contextmanager
from distutils.util import strtobool
from functools import wraps
from logging import DEBUG, getLogger
from json import loads
from traceback import format_exc

from flask import current_app, g, request
from microcosm.api import defaults
from microcosm_flask.errors import (
    extract_context,
    extract_error_message,
    extract_include_stack_trace,
    extract_status_code,
)
from microcosm_logging.timing import elapsed_time


AuditOptions = namedtuple("AuditOptions", [
    "include_request_body",
    "include_response_body",
])


SKIP_LOGGING = "_microcosm_flask_skip_audit_logging"


def skip_logging(func):
    """
    Decorate a function so logging will be skipped.

    """
    setattr(func, SKIP_LOGGING, True)
    return func


def should_skip_logging(func):
    """
    Should we skip logging for this handler?

    """
    return getattr(func, SKIP_LOGGING, False)


@contextmanager
def logging_levels():
    """
    Context manager to conditionally set logging levels.

    Supports setting per-request debug logging using the `X-Request-Debug` header.

    """
    enabled = strtobool(request.headers.get("x-request-debug", "false"))
    level = None
    try:
        if enabled:
            level = getLogger().getEffectiveLevel()
            getLogger().setLevel(DEBUG)
        yield
    finally:
        if enabled:
            getLogger().setLevel(level)


def audit(func):
    """
    Record a Flask route function in the audit log.

    Generates a JSON record in the Flask log for every request.

    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        options = AuditOptions(
            include_request_body=True,
            include_response_body=True,
        )
        with logging_levels():
            return _audit_request(options, func, None, *args, **kwargs)

    return wrapper


class RequestInfo(object):
    """
    Capture of key information for requests.

    """
    def __init__(self, options, func, request_context):
        self.options = options
        self.operation = request.endpoint
        self.func = func.__name__
        self.method = request.method
        self.request_context = request_context
        self.timing = dict()

        self.error = None
        self.stack_trace = None
        self.request_body = None
        self.response_body = None
        self.status_code = None
        self.success = None

    def to_dict(self):
        dct = dict(
            operation=self.operation,
            func=self.func,
            method=self.method,
            **self.timing
        )

        if self.request_context is not None:
            dct.update(self.request_context())

        if self.success is True:
            dct.update(
                success=self.success,
                status_code=self.status_code,
            )
        if self.success is False:
            dct.update(
                success=self.success,
                message=extract_error_message(self.error)[:2048],
                context=extract_context(self.error),
                stack_trace=self.stack_trace,
                status_code=self.status_code,
            )

        self.post_process_request_body(dct)
        self.post_process_response_body(dct)

        return dct

    def log(self, logger):
        if self.status_code == 500:
            # something actually went wrong; investigate
            logger.warning(self.to_dict())
        else:
            # usually log at INFO; a raised exception can be an error or expected behavior (e.g. 404)
            logger.info(self.to_dict())

    def capture_request(self):
        if not current_app.debug:
            # only capture request body on debug
            return

        if not self.options.include_request_body:
            # only capture request body if requested
            return

        if not request.get_json(force=True, silent=True):
            # only capture request body if json
            return

        self.request_body = request.get_json(force=True)

    def capture_response(self, response):
        self.success = True

        body, self.status_code = parse_response(response)

        if not current_app.debug:
            # only capture responsebody on debug
            return

        if not self.options.include_response_body:
            # only capture response body if requested
            return

        if not body:
            # only capture request body if there is one
            return

        try:
            self.response_body = loads(body)
        except (TypeError, ValueError):
            # not json
            pass

    def capture_error(self, error):
        self.error = error
        self.status_code = extract_status_code(error)
        self.success = 0 < self.status_code < 400
        include_stack_trace = extract_include_stack_trace(error)
        self.stack_trace = format_exc(limit=10) if (not self.success and include_stack_trace) else None

    def post_process_request_body(self, dct):
        if g.get("hide_body") or not self.request_body:
            return

        for name, new_name in g.get("show_request_fields", {}).items():
            try:
                value = self.request_body.pop(name)
                self.request_body[new_name] = value
            except KeyError:
                pass

        for field in g.get("hide_request_fields", []):
            try:
                del self.request_body[field]
            except KeyError:
                pass

        dct.update(
            request_body=self.request_body,
        )

    def post_process_response_body(self, dct):
        if g.get("hide_body") or not self.response_body:
            return

        for name, new_name in g.get("show_response_fields", {}).items():
            try:
                value = self.response_body.pop(name)
                self.response_body[new_name] = value
            except KeyError:
                pass

        for field in g.get("hide_response_fields", []):
            try:
                del self.response_body[field]
            except KeyError:
                pass

        dct.update(
            response_body=self.response_body,
        )


def _audit_request(options, func, request_context, *args, **kwargs):  # noqa: C901
    """
    Run a request function under audit.

    """
    logger = getLogger("audit")

    request_info = RequestInfo(options, func, request_context)
    response = None

    request_info.capture_request()
    try:
        # process the request
        with elapsed_time(request_info.timing):
            response = func(*args, **kwargs)
    except Exception as error:
        request_info.capture_error(error)
        raise
    else:
        request_info.capture_response(response)
        return response
    finally:
        if not should_skip_logging(func):
            request_info.log(logger)


def parse_response(response):
    """
    Parse a Flask response into a body and status code.

    The returned value from a Flask view could be:
        * a tuple of (response, status) or (response, status, headers)
        * a Response object
        * a string
    """
    if isinstance(response, tuple) and len(response) > 1:
        return response[0], response[1]
    try:
        return response.data, response.status_code
    except AttributeError:
        return response, 200


@defaults(
    include_request_body=True,
    include_response_body=True,
)
def configure_audit_decorator(graph):
    """
    Configure the audit decorator.

    Example Usage:

        @graph.audit
        def login(username, password):
            ...
    """
    include_request_body = graph.config.audit.include_request_body
    include_response_body = graph.config.audit.include_response_body

    def _audit(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            options = AuditOptions(
                include_request_body=include_request_body,
                include_response_body=include_response_body,
            )
            return _audit_request(options, func, graph.request_context,  *args, **kwargs)
        return wrapper
    return _audit



"""
Push resource definitions to an output destination.

"""
from json import dumps
from logging import getLogger
from sys import stdout

import requests
from requests.exceptions import ConnectionError, HTTPError
from six.moves.urllib.parse import urlparse, urlunparse
from yaml import safe_dump_all


logger = getLogger("sync.push")


def push_yaml(inputs, destination):
    """
    Write inputs to destination as YAML.

    :param inputs: an iterable of (href, resource) pairs
    :param destination: a writable file-like object

    """
    safe_dump_all(({href: resource} for href, resource in inputs), destination)


def push_json(inputs, base_url, batch_size, enable_sessions=False,  keep_instance_path=False, max_attempts=2):
    """
    Write inputs to remote URL as JSON.

    Future implementations could perform paginated PATCH requests to a collection URI
    if our conventions support it.

    """
    # either use a session or use plain requests
    session_factory = requests.Session if enable_sessions else lambda: requests

    session = session_factory()
    for uri, resources in iter_json_batches(inputs, base_url, batch_size, keep_instance_path):
        # retry on connection failures
        last_error = None
        for attempt in range(max_attempts):
            try:
                if batch_size == 1:
                    push_resource_json(session, uri, resources[0])
                else:
                    push_resource_json_batch(session, uri, resources)
            except ConnectionError as error:
                logger.info("Connection error for uri: {}: {}".format(uri, error))
                # on connection failure, recreate the session
                session = session_factory()
                last_error = error
                continue
            except HTTPError as error:
                if error.response.status_code in (504, 502):
                    logger.info("HTTP error for uri: {}: {}".format(uri, error))
                    # on connection failure, recreate the session
                    session = session_factory()
                    last_error = error
                    continue
                raise
            else:
                break
        else:
            # If reached here, all attempts were unsuccessful - raise last error encountered
            raise last_error


def iter_json_batches(inputs, base_url, batch_size, keep_instance_path):
    parsed_base_url = urlparse(base_url)

    current_uri = None
    current_batch = []

    for href, resource in inputs:
        # Skip over links-only (discovery) resources
        if resource.keys() == ["_links"]:
            continue

        # Inject the base URL's scheme and netloc; `urljoin` should do exactly this operation,
        # but actually won't if the right-hand-side term defines its own netloc
        parsed_href = urlparse(href)
        uri = urlunparse(parsed_href._replace(
            scheme=parsed_base_url.scheme,
            netloc=parsed_base_url.netloc,
        ))

        if batch_size == 1:
            yield (uri, [resource])
        else:
            # batch handling
            if keep_instance_path:
                collection_uri = uri.rsplit("?", 1)[0]
            else:
                collection_uri = uri.rsplit("/", 1)[0]

            if any((
                    current_uri is not None and current_uri != collection_uri,
                    len(current_batch) >= batch_size,
            )):
                yield (current_uri, current_batch)
                current_batch = []

            current_uri = collection_uri
            current_batch.append(resource)

    if current_batch:
        yield (current_uri, current_batch)


def push_resource_json(session, uri, resource):
    """
    Push a single resource as JSON to a URI.

    Assumes that the backend supports a replace/put convention.

    """
    logger.debug("Pushing resource for {}".format(uri))

    response = session.put(
        uri,
        data=dumps(resource),
        headers={"Content-Type": "application/json"},
    )
    try:
        response.raise_for_status()
    except:
        logger.warning("Unable to replace {}".format(
            uri,
        ))
        raise


def push_resource_json_batch(session, uri, resources):
    """
    Push a single resource as JSON to a URI.

    Assumes that the backend supports a replace/put convention.

    """
    logger.debug("Pushing resource batch of size {} for {}".format(len(resources), uri))

    response = session.patch(
        uri,
        data=dumps(dict(
            items=resources,
        )),
        headers={"Content-Type": "application/json"},
    )
    try:
        response.raise_for_status()
    except:
        logger.warning("Unable to replace {}".format(
            uri,
        ))
        raise


def push(args, inputs):
    """
    Push content to a destination.

    If the destination is "-", YAML is written to stdout.
    If the destination has a http prefix, JSON is written to a URL.
    Otherwise, YAML is written to a local file.

    """
    logger.info("Pushing resources to: {}".format(args.output))

    if args.output == "-":
        push_yaml(inputs, stdout)
    elif args.output.startswith("http"):
        push_json(inputs, args.output, args.batch_size, args.enable_sessions, args.keep_instance_path)
    else:
        with open(args.output, "w") as file_:
            push_yaml(inputs, file_)



"""
Swagger (OpenAPI) convention.

Exposes swagger definitions for matching operations.

"""
from flask import g

from microcosm.api import defaults
from microcosm_flask.conventions.base import Convention
from microcosm_flask.conventions.encoding import make_response
from microcosm_flask.conventions.registry import iter_endpoints
from microcosm_flask.namespaces import Namespace
from microcosm_flask.operations import Operation
from microcosm_flask.routing import make_path
from microcosm_flask.swagger.definitions import build_swagger


class SwaggerConvention(Convention):

    @property
    def matching_operations(self):
        return {
            Operation.from_name(operation_name)
            for operation_name in self.graph.config.swagger_convention.operations
        }

    def find_matching_endpoints(self, swagger_ns):
        """
        Compute current matching endpoints.

        Evaluated as a property to defer evaluation.

        """
        def match_func(operation, ns, rule):
            # only expose endpoints that have the correct path prefix and operation
            return (
                rule.rule.startswith(make_path(self.graph, swagger_ns.path)) and
                operation in self.matching_operations
            )

        return list(iter_endpoints(self.graph, match_func))

    def configure_discover(self, ns, definition):
        """
        Register a swagger endpoint for a set of operations.

        """
        @self.add_route(ns.singleton_path, Operation.Discover, ns)
        def discover():
            swagger = build_swagger(self.graph, ns, self.find_matching_endpoints(ns))
            g.hide_body = True
            return make_response(swagger)


@defaults(
    name="swagger",
    operations=[
        "create",
        "create_collection",
        "create_for",
        "delete",
        "replace",
        "replace_for",
        "retrieve",
        "retrieve_for",
        "search",
        "search_for",
        "update",
        "update_batch",
        "update_for",
        "upload",
        "upload_for",
    ],
    path_prefix="",
    version="",
)
def configure_swagger(graph):
    """
    Build a singleton endpoint that provides swagger definitions for all operations.

    """
    ns = Namespace(
        path=graph.config.swagger_convention.path_prefix,
        subject=graph.config.swagger_convention.name,
        version=graph.config.swagger_convention.version,
    )
    convention = SwaggerConvention(graph)
    convention.configure(ns, discover=tuple())
    return ns.subject



"""
Custom fields.

"""
from microcosm_flask.fields.enum_field import EnumField  # noqa: F401
from microcosm_flask.fields.language_field import LanguageField  # noqa: F401
from microcosm_flask.fields.query_string_list import QueryStringList  # noqa: F401
from microcosm_flask.fields.timestamp_field import TimestampField  # noqa: F401
from microcosm_flask.fields.uri_field import URIField  # noqa: F401



"""
Naming conventions.

"""
from inspect import isclass

from inflection import underscore
from six import string_types


def name_for(obj):
    """
    Get a name for something.

    Allows overriding of default names using the `__alias__` attribute.

    """
    if isinstance(obj, string_types):
        return obj

    cls = obj if isclass(obj) else obj.__class__

    if hasattr(cls, "__alias__"):
        return underscore(cls.__alias__)
    else:
        return underscore(cls.__name__)


def collection_path_for(name):
    """
    Get a path for a collection of things.

    """
    return "/{}".format(
        name_for(name),
    )


def singleton_path_for(name):
    """
    Get a path for a singleton thing.

    """
    return "/{}".format(
        name_for(name),
    )


def instance_path_for(name, identifier_type):
    """
    Get a path for thing.

    """
    return "/{}/<{}:{}_id>".format(
        name_for(name),
        identifier_type,
        name_for(name),
    )


def alias_path_for(name):
    """
    Get a path for an alias to a thing

    """
    return "/{}/<{}_name>".format(
        name_for(name),
        name_for(name),
    )


def relation_path_for(from_name, to_name, identifier_type):
    """
    Get a path relating a thing to another.

    """
    return "/{}/<{}:{}_id>/{}".format(
        name_for(from_name),
        identifier_type,
        name_for(from_name),
        name_for(to_name),
    )



"""
Expose/manage logging levels.

"""
from collections import namedtuple
from logging import PlaceHolder, root

from marshmallow import fields, Schema
from microcosm.api import defaults

from microcosm_flask.conventions.base import Convention, EndpointDefinition
from microcosm_flask.conventions.encoding import dump_response_data
from microcosm_flask.conventions.registry import response
from microcosm_flask.namespaces import Namespace
from microcosm_flask.operations import Operation


Logger = namedtuple("Logger", ["name", "level", "logger", "children"])


def dehumanize_level(value):
    return {
        "DEBUG": 10,
        "INFO": 20,
        "WARNING": 30,
        "ERROR": 30,
        "CRITICAL": 50,
    }[value]


def humanize_level(value):
    try:
        return {
            0: "NOTSET",
            10: "DEBUG",
            20: "INFO",
            30: "WARNING",
            40: "ERROR",
            50: "CRITICAL",
        }[int(value)]
    except:
        return value


def make_logger_node(name, logger, parent=None):
    if isinstance(logger, PlaceHolder):
        # some loggers are instances of logging.PlaceHolder
        level = parent.level
    else:
        level = humanize_level(logger.getEffectiveLevel())

    node = Logger(name, level, logger, [])
    if parent is not None:
        parent.children.append(node)
    return node


def build_logger_tree():
    """
    Build a DFS tree representing the logger layout.

    Adapted with much appreciation from: https://github.com/brandon-rhodes/logging_tree

    """
    cache = {}
    tree = make_logger_node("", root)
    for name, logger in sorted(root.manager.loggerDict.items()):
        if "." in name:
            parent_name = ".".join(name.split(".")[:-1])
            parent = cache[parent_name]
        else:
            parent = tree

        cache[name] = make_logger_node(name, logger, parent)
    return tree


class LoggerSchema(Schema):
    name = fields.String(required=True)
    level = fields.String(required=True)
    children = fields.List(fields.Nested("LoggerSchema"), required=True)


class LoggingLevelConvention(Convention):
    """
    Convention that exposing logging configuration.

    Includes a `GET` request to retrieve the current logging tree and a `PATCH` request
    to (temporarily) change logging levels.

    """
    def __init__(self, graph):
        super(LoggingLevelConvention, self).__init__(graph)
        self.max_duration = float(graph.config.logging_level_convention.max_duration)

    def configure_retrieve(self, ns, definition):
        response_schema = definition.response_schema

        @self.add_route(ns.singleton_path, Operation.Retrieve, ns)
        @response(response_schema)
        def retrieve():
            logger_tree = build_logger_tree()
            return dump_response_data(response_schema, logger_tree)


@defaults(
    max_duration=300.0,
)
def configure_logging_level(graph):
    ns = Namespace(
        subject="logging_level",
    )

    convention = LoggingLevelConvention(graph)
    convention.configure(
        ns,
        retrieve=EndpointDefinition(
            response_schema=LoggerSchema(),
        ),
    )
    return ns
